#!/usr/bin/env python3
# encoding: UTF-8

import re
import socket
import sys
import xml.etree.ElementTree as ET

if sys.version_info.major == 3:
    import urllib.request as urllibreq
    import urllib.parse as urllibparse
else:
    import urllib2 as urllibreq
    import urlparse as urllibparse

import logging
import json

SSDP_BROADCAST_PORT = 1900
SSDP_BROADCAST_ADDR = "239.255.255.250"

SSDP_BROADCAST_PARAMS = [
    "M-SEARCH * HTTP/1.1",
    "HOST: {0}:{1}".format(SSDP_BROADCAST_ADDR, SSDP_BROADCAST_PORT),
    "MAN: \"ssdp:discover\"", "MX: 10", "ST: ssdp:all", "", ""]
SSDP_BROADCAST_MSG = "\r\n".join(SSDP_BROADCAST_PARAMS)

UPNP_DEFAULT_SERVICE_TYPE = "urn:schemas-upnp-org:service:AVTransport:1"


def register_device(location_url):

    xml = urllibreq.urlopen(location_url).read().decode("UTF-8")
    xml = re.sub(" xmlns=\"[^\"]+\"", "", xml, count=1)
    info = ET.fromstring(xml)

    location = urllibparse.urlparse(location_url)
    hostname = location.hostname

    friendly_name = info.find("./device/friendlyName").text

    try:
        path = info.find(
            "./device/serviceList/service/"
            "[serviceType='{0}']/controlURL".format(
                UPNP_DEFAULT_SERVICE_TYPE
            )
        ).text
        action_url = urllibparse.urljoin(location_url, path)
    except AttributeError:
        action_url = None

    device = {
        "location": location_url,
        "hostname": hostname,
        "friendly_name": friendly_name,
        "action_url": action_url,
        "st": UPNP_DEFAULT_SERVICE_TYPE
    }

    logging.debug(
        "Device registered: {}".format(
            json.dumps({
                "device_xml": xml,
                "device_info": device
            })
        )
    )

    return device


def get_devices(timeout=3.0):

    logging.debug("Configuring broadcast message")
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
    s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 4)
    s.bind(("", SSDP_BROADCAST_PORT + 10))

    logging.debug("Sending broadcast message")
    s.sendto(SSDP_BROADCAST_MSG.encode("UTF-8"), (SSDP_BROADCAST_ADDR,
                                                  SSDP_BROADCAST_PORT))

    logging.debug("Waiting for devices ({} seconds)".format(timeout))
    s.settimeout(timeout)

    devices = []
    while True:

        try:
            data, addr = s.recvfrom(1024)
        except socket.timeout:
            break

        try:
            info = [a.split(":", 1)
                    for a in data.decode("UTF-8").split("\r\n")[1:]]
            device = dict([(a[0].strip().lower(), a[1].strip())
                           for a in info if len(a) >= 2])
            devices.append(device)
            logging.debug(
                "Device broadcast response: {}".format(
                    json.dumps({
                        "broadcast_message_raw": data.decode("UTF-8"),
                        "broadcast_message_info": device
                    })
                )
            )
        except Exception:
            pass

    devices_urls = [dev["location"]
                    for dev in devices if "AVTransport" in dev["st"]]
    devices = [register_device(location_url) for location_url in devices_urls]

    return devices


if __name__ == "__main__":

    timeout = int(sys.argv[1]) if len(sys.argv) >= 2 else 5

    devices = get_devices(timeout)

    for i, device in enumerate(devices, 1):
        print("Device {0}:\n{1}\n\n".format(i, json.dumps(device, indent=4)))



#!/usr/bin/env python3

from __future__ import print_function

import argparse
import json
import os
import sys
import datetime

from . import devices, dlna, streaming

import logging


def get_subtitle(file_video):

    video, extension = os.path.splitext(file_video)

    file_subtitle = "{0}.srt".format(video)

    if not os.path.exists(file_subtitle):
        return None
    return file_subtitle


def list_devices(args):

    logging.info("Scanning devices...")
    my_devices = devices.get_devices(args.timeout)
    logging.info("Number of devices found: {}".format(len(my_devices)))

    for i, device in enumerate(my_devices, 1):
        print("Device {0}:\n{1}\n\n".format(i, json.dumps(device, indent=4)))


def play(args):

    logging.info("Starting to play")

    # Get video and subtitle file names

    files = {"file_video": args.file_video}

    if args.use_subtitle:

        if not args.file_subtitle:
            args.file_subtitle = get_subtitle(args.file_video)

        if args.file_subtitle:
            files["file_subtitle"] = args.file_subtitle

    logging.info("Media files: {}".format(json.dumps(files)))

    # Select device to play
    logging.info("Selecting device to play")

    device = None

    if args.device_url:
        logging.info("Select device by URL")
        device = devices.register_device(args.device_url)
    else:
        my_devices = devices.get_devices(args.timeout)

        if len(my_devices) > 0:
            if args.device_query:
                logging.info("Select device by query")
                device = [
                    device for device in my_devices
                    if args.device_query.lower() in str(device).lower()][0]
            else:
                logging.info("Select first device")
                device = my_devices[0]

    if not device:
        sys.exit("No devices found.")

    logging.info("Device selected: {}".format(json.dumps(device)))

    # Configure streaming server
    logging.info("Configuring streaming server")

    target_ip = device["hostname"]
    if args.local_host:
        serve_ip = args.local_host
    else:
        serve_ip = streaming.get_serve_ip(target_ip)
    files_urls = streaming.start_server(files, serve_ip)

    logging.info("Streaming server ready")

    # Play the video through DLNA protocol
    logging.info("Sending play command")
    dlna.play(files_urls, device)


def run():

    parser = argparse.ArgumentParser(
        description="A minimal UPnP/DLNA media streamer.")
    parser.add_argument("-t", "--timeout", type=float, default=5)
    parser.add_argument("-b", "--debug",
                        dest="debug_activated", action="store_true")
    subparsers = parser.add_subparsers(dest="subparser_name")

    p_list = subparsers.add_parser('list')
    p_list.set_defaults(func=list_devices)

    p_play = subparsers.add_parser('play')
    p_play.add_argument("-d", "--device", dest="device_url")
    p_play.add_argument("-H", "--host", dest="local_host")
    p_play.add_argument("-q", "--query-device", dest="device_query")
    p_play.add_argument("-s", "--subtitle", dest="file_subtitle")
    p_play.add_argument("-n", "--no-subtitle",
                        dest="use_subtitle", action="store_false")
    p_play.add_argument("file_video")
    p_play.set_defaults(func=play)

    args = parser.parse_args()

    log_filename = "nanodlna-{}.log".format(
        datetime.datetime.today().strftime("%Y-%m-%d_%H-%M-%S")
    )

    logging.basicConfig(
        filename=log_filename,
        filemode="w",
        level=logging.INFO,
        format="[ %(asctime)s ] %(levelname)s : %(message)s"
    )
    print("nano-dlna log will be saved here: {}".format(log_filename))

    if args.debug_activated:
        logging.getLogger().setLevel(logging.DEBUG)

    args.func(args)


if __name__ == "__main__":

    run()



#!/usr/bin/env python3
# encoding: UTF-8

import os
import socket
import threading
import unicodedata
import re

from twisted.internet import reactor
from twisted.web.resource import Resource
from twisted.web.server import Site
from twisted.web.static import File

import logging
import json

# from twisted.python import log


def normalize_file_name(value):
    value = unicodedata\
        .normalize("NFKD", value)\
        .encode("ascii", "ignore")\
        .decode("ascii")
    value = re.sub(r"[^\.\w\s-]", "", value.lower())
    value = re.sub(r"[-\s]+", "-", value).strip("-_")
    return value


def set_files(files, serve_ip, serve_port):

    logging.debug("Setting streaming files: {}".format(
        json.dumps({
            "files": files,
            "serve_ip": serve_ip,
            "serve_port": serve_port
        })
    ))

    files_index = {file_key: (normalize_file_name(os.path.basename(file_path)),
                              os.path.abspath(file_path),
                              os.path.dirname(os.path.abspath(file_path)))
                   for file_key, file_path in files.items()}

    files_serve = {file_name: file_path
                   for file_name, file_path, file_dir in files_index.values()}

    files_urls = {
        file_key: "http://{0}:{1}/{2}/{3}".format(
            serve_ip, serve_port, file_key, file_name)
        for file_key, (file_name, file_path, file_dir)
        in files_index.items()}

    logging.debug("Streaming files information: {}".format(
        json.dumps({
            "files_index": files_index,
            "files_serve": files_serve,
            "files_urls": files_urls
        })
    ))

    return files_index, files_serve, files_urls


def start_server(files, serve_ip, serve_port=9000):

    # import sys
    # log.startLogging(sys.stdout)

    logging.debug("Starting to create streaming server")

    files_index, files_serve, files_urls = set_files(
        files, serve_ip, serve_port)

    logging.debug("Adding files to HTTP server")
    root = Resource()
    for file_key, (file_name, file_path, file_dir) in files_index.items():
        root.putChild(file_key.encode("utf-8"), Resource())
        root.children[file_key.encode("utf-8")].putChild(
            file_name.encode("utf-8"), File(file_path))

    logging.debug("Starting to listen messages in HTTP server")
    reactor.listenTCP(serve_port, Site(root))
    threading.Thread(
        target=reactor.run, kwargs={"installSignalHandlers": False}).start()

    return files_urls


def get_serve_ip(target_ip, target_port=80):
    logging.debug("Identifying server IP")
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect((target_ip, target_port))
    serve_ip = s.getsockname()[0]
    s.close()
    logging.debug("Server IP identified: {}".format(serve_ip))
    return serve_ip


if __name__ == "__main__":

    import sys

    files = {"file_{0}".format(i): file_path for i,
             file_path in enumerate(sys.argv[1:], 1)}
    print(files)

    files_urls = start_server(files, "localhost")
    print(files_urls)



#!/usr/bin/env python

import sys

from setuptools import setup


_INSTALL_REQUIRES = [
    'Twisted>=16.2.0',
]
if sys.version_info.major == 2:
    _INSTALL_REQUIRES.append("service_identity>=16.0.0")

setup(
    name='nanodlna',
    version='0.2.0',
    description='A minimal UPnP/DLNA media streamer',
    long_description="""nano-dlna is a command line tool that allows you to
 play a local video file in your TV (or any other DLNA compatible device)""",
    author='Gabriel Magno',
    author_email='gabrielmagno1@gmail.com',
    url='https://github.com/gabrielmagno/nano-dlna',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Multimedia :: Video',
        'Topic :: Utilities'
    ],
    zip_safe=True,
    packages=['nanodlna'],
    package_dir={'nanodlna': 'nanodlna'},
    package_data={'nanodlna': ['templates/*.xml']},
    entry_points={
        'console_scripts': [
            'nanodlna = nanodlna.cli:run',
        ]
    },
    install_requires=_INSTALL_REQUIRES
)



# -*- coding: utf-8 -*-

__title__ = 'nanodlna'
__version__ = '0.2.0'
__short_version__ = '.'.join(__version__.split('.')[:2])
__author__ = 'Gabriel Magno'
__license__ = 'MIT'
__copyright__ = 'Copyright 2016, Gabriel Magno'



#!/usr/bin/env python3
# encoding: UTF-8

import os
import pkgutil
import sys
from xml.sax.saxutils import escape as xmlescape

if sys.version_info.major == 3:
    import urllib.request as urllibreq
else:
    import urllib2 as urllibreq

import traceback
import logging
import json


def send_dlna_action(device, data, action):

    logging.debug("Sending DLNA Action: {}".format(
        json.dumps({
            "action": action,
            "device": device,
            "data": data
        })
    ))

    action_data = pkgutil.get_data(
        "nanodlna", "templates/action-{0}.xml".format(action)).decode("UTF-8")
    action_data = action_data.format(**data).encode("UTF-8")

    headers = {
        "Content-Type": "text/xml; charset=\"utf-8\"",
        "Content-Length": "{0}".format(len(action_data)),
        "Connection": "close",
        "SOAPACTION": "\"{0}#{1}\"".format(device["st"], action)
    }

    logging.debug("Sending DLNA Request: {}".format(
        json.dumps({
            "url": device["action_url"],
            "data": action_data.decode("UTF-8"),
            "headers": headers
        })
    ))

    try:
        request = urllibreq.Request(device["action_url"], action_data, headers)
        urllibreq.urlopen(request)
        logging.debug("Request sent")
    except Exception:
        logging.error("Unknown error sending request: {}".format(
            json.dumps({
                "url": device["action_url"],
                "data": action_data.decode("UTF-8"),
                "headers": headers,
                "error": traceback.format_exc()
            })
        ))


def play(files_urls, device):

    logging.debug("Starting to play: {}".format(
        json.dumps({
            "files_urls": files_urls,
            "device": device
        })
    ))

    video_data = {
        "uri_video": files_urls["file_video"],
        "type_video": os.path.splitext(files_urls["file_video"])[1][1:],
    }

    if "file_subtitle" in files_urls and files_urls["file_subtitle"]:

        video_data.update({
            "uri_sub": files_urls["file_subtitle"],
            "type_sub": os.path.splitext(files_urls["file_subtitle"])[1][1:]
        })

        metadata = pkgutil.get_data(
            "nanodlna",
            "templates/metadata-video_subtitle.xml").decode("UTF-8")
        video_data["metadata"] = xmlescape(metadata.format(**video_data))

    else:
        video_data["metadata"] = ""

    logging.debug("Created video data: {}".format(json.dumps(video_data)))

    logging.debug("Setting Video URI")
    send_dlna_action(device, video_data, "SetAVTransportURI")
    logging.debug("Playing video")
    send_dlna_action(device, video_data, "Play")
